import os, shutil
import surfinBH
import h5py

def test_download_links():
    """ Tests that the download links for fit data are working.
    """

    # dir to download data to
    out_dir = 'test/download_data'

    # remove out_dir if it already exists and make a new one
    if os.path.exists(out_dir):
        shutil.rmtree(out_dir)
    os.system('mkdir -p %s'%out_dir)

    # List of all available fits
    fit_names = surfinBH.fits_collection.keys()
    for name in fit_names:
        surfinBH.DownloadData(name=name, data_dir=out_dir)

        # allow for both naming formats surfinBH7dq2 and NRSur7dq4Remnant
        if 'surfinBH' in name:
            name_tag = name.split('surfinBH')[-1]
        else:
            name_tag = name.split('NRSur')[-1].split('Remnant')[0]

        # check that it has the right name
        assert(os.path.isfile('%s/fit_%s.h5'%(out_dir, name_tag)))
        # check that the fit_name matches with the name in the attributes
        # of h5 file.
        h5file = h5py.File('%s/fit_%s.h5'%(out_dir, name_tag), 'r')
        assert(name_tag == h5file.attrs['name'].decode('utf-8'))
        h5file.close()



#!/usr/bin/python
import numpy as np
from math import pi, ceil
import warnings
from random import sample as rand_sample

from sklearn import gaussian_process
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn import linear_model

#######################     GPR fit functions       #######################

# These are the things we need to save from a GPR fit in order to reproduce it
# NOTE: If you get errors like "AttributeError: 'GaussianProcessRegressor'
# object has no attribute -----", try adding that attribute to
# GPR_SAVE_ATTRS_DICT
GPR_SAVE_ATTRS_DICT = ['kernel_', 'X_train_', 'alpha_', '_y_train_mean', 'L_']

# These are the things we need to save from a LinearRegression fit in order
# to reproduce it
# NOTE: If you get errors like "AttributeError: 'LinearRegression'
# object has no attribute -----", try adding that attribute to
# LINREG_SAVE_ATTRS_DICT
LINREG_SAVE_ATTRS_DICT = ['coef_', 'intercept_']

# -------------------------------------------------------------------------
class GPRPredictor:
    def __init__(self, res, **kwargs):
        """
        Class to evaluate GPR fits constructed by GPRFitter class in
        pySurrogate/fit_gpr.py
        """

        self.data_mean = res['data_mean']
        self.data_std = res['data_std']

        # load GPR fit
        self.GPR_obj = GaussianProcessRegressor()
        GPR_params =  res['GPR_params']
        self._set_gpr_params(self.GPR_obj, GPR_params)

        # load LinearRegression fit
        lin_reg_params = res['lin_reg_params']
        if lin_reg_params is not None:
            self.linearModel = linear_model.LinearRegression()
            self._set_lin_reg_params(self.linearModel, lin_reg_params)
        else:
            self.linearModel = None

    def _set_kernel_params(self, kernel_params):
        """ Recursively sets paramters for a kernel and returns the final
            kernel.
        """

        # copy so as to not overwrite
        kernel_params = kernel_params.copy()

        # initialize kernel
        name = kernel_params['name']
        kernel = getattr(gaussian_process.kernels, name)
        del kernel_params['name']

        params = {}
        for key in kernel_params.keys():
            if type(kernel_params[key]) == dict:
                # recursively set kernels
                params[key] = self._set_kernel_params(kernel_params[key])
            else:
                params[key] = kernel_params[key]

        if name == 'Sum' or name == 'Product':
            kernel = kernel(params['k1'], params['k2'])
        else:
            kernel = kernel()

        kernel.set_params(**params)
        return kernel


    def _set_gpr_params(self, gp_obj, gp_params):
        """ Sets the fitted hyperparameter for a GPR object.
            This can be used to load a previously constructed fit.

            NOTE: If you get errors like:
            "AttributeError: 'GaussianProcessRegressor' object has
            no attribute -----",
            try adding that attribute to GPR_SAVE_ATTRS_DICT
        """
        for attr in GPR_SAVE_ATTRS_DICT:
            if attr == 'kernel_':
                param = self._set_kernel_params(gp_params[attr])
            else:
                param = gp_params[attr]
            setattr(gp_obj, attr, param)


    def _set_lin_reg_params(self, lr_obj, lr_params):
        """ Sets the fitted parameters for a LinearRegression object.
            This can be used to load a previously constructed fit.

            NOTE: If you get errors like:
            "AttributeError: 'LinearRegression' object has
            no attribute -----",
            try adding that attribute to LINREG_SAVE_ATTRS_DICT
        """
        for attr in LINREG_SAVE_ATTRS_DICT:
            param = lr_params[attr]
            setattr(lr_obj, attr, param)


    def _reconstruct_normalized_data(self, data_normed, data_normed_err):
        """
        The inverse operation of 'GPRFitter._normalize()'
        Returns the reconstructed data and error estimate.
        """
        return data_normed * self.data_std + self.data_mean, \
            data_normed_err * self.data_std

    def GPR_predict(self, x, estimate_err=False):
        """
        Evaluates a GPR fit.
        First evalutates the GPR fit to get the prediction for the normalized
        data. Then reconstructs the un-normalized data.
        Finally adds the linear model fit if it was done in GPRFitter.
        """

        # Get fit prediction and error estimate for normalized data
        fit_res = self.GPR_obj.predict(x, return_cov=estimate_err)

        if estimate_err:
            y_normalized_pred, cov_normalized_pred = fit_res
            err_normalized_pred = np.sqrt(cov_normalized_pred.flatten())
        else:
            y_normalized_pred = fit_res
            err_normalized_pred = fit_res * 0

        # Reconstruct to get un-normalized prediction
        y_pred, err_pred = self._reconstruct_normalized_data( \
                                            y_normalized_pred, \
                                            err_normalized_pred)

        if self.linearModel is not None:
            # Add the linear prediction that was subtracted before
            # doing the fit
            y_pred = y_pred + self.linearModel.predict(x)


        val_dict = {
            'y': y_pred,
        }

        if estimate_err:
            val_dict['y_gprErr'] = err_pred

        return val_dict


#######################     Greedt fit functions       #######################
class BasisFunction:
    def __init__(self, name, func, minX, maxX):
        """name: A name (for example, 'cosine')
        func: f(n, x) gives the n'th basis function evaluated at x
            for example, lambda n, x: np.cos(n*x)
        minX, maxX: The default domain (for example, minX=0, maxX=pi)
            If the user's domain is different, linearly maps to the default.
            Can give None, in which case the domain is not mapped.
        """
        self.name = name
        self.func = func
        self.minX = minX
        if len([tmp for tmp in [minX, maxX] if tmp is None]) == 1:
            raise Exception('Must have both or neither minX=None, maxX=None')
        if self.minX is None:
            self.width = None
        else:
            self.width = maxX - minX

    def __call__(self, n, x):
        return self.func(n, x)

    def mappedFunc(self, n, minX, maxX):
        if self.minX is None:
            return lambda x: self.func(n, x)
        def newFunc(x):
            return self(n, self.minX + (x-minX)*self.width/(maxX-minX))
        return newFunc

## Define some commonly used BasisFunctions
_polynomial = BasisFunction(
        'polynomial',
        lambda n, x: 1. + 0.*x if n==0 else x**n,
        -1, 1)

_complexPeriodic = BasisFunction(
        'complexPeriodic',
        lambda n, x: np.exp(1.j*n*x/2) if n%2==0 else np.exp(-1.j*(n+1)*x/2),
        0, 2*pi)

_periodic = BasisFunction(
        'periodic',
        lambda n, x: np.cos(n*x/2) if n%2==0 else np.sin((n+1)*x/2),
        0, 2*pi)

_chebychev = BasisFunction(
        'chebyshev',
        lambda n, x: np.polynomial.Chebyshev([0]*n + [1])(x),
        -1, 1)

bf_dict = {
            'polynomial': _polynomial,
            'complexPeriodic': _complexPeriodic,
            'periodic': _periodic,
            'chebyshev': _chebychev,
          }

def basisFunction(typeStr, n, minVal, maxVal, BF_dict=bf_dict):
    return BF_dict[typeStr].mappedFunc(n, minVal, maxVal)


#######################   Fit evaluator functions   #######################
def getFitEvaluator(res):
    if res is None:
        return None

    def greedyfitEvaluator(x):
        basisFuncEvals = np.array([
                np.prod(
                  np.array([
                    basisFunction(
                      res['bfTypes'][j],
                      orders[j],
                      res['minVals'][j],
                      res['maxVals'][j])(x[j])
                    for j in range(len(x))]), 0)
                for orders in res['bfOrders']])
        return res['coefs'].dot(basisFuncEvals)

    def gprfitEvaluator(x):
        GPR_eval  = gpr_predictObject.GPR_predict([x])
        return GPR_eval['y'][0] # This probaly needs to change if len(x) != 1

    if 'fitType' in res and res['fitType'] == 'GPR':
        gpr_predictObject = GPRPredictor(res)
        return gprfitEvaluator
    else:
        return greedyfitEvaluator


# Evalutates the GPR error estimation function.
# TODO: Currently this just reevaluates the fit, consider incorporating into
# getFitEvaluator() somehow
def getGPRErrorEvaluator(res):

    def gprErrEval(x):
        GPR_eval  = gpr_predictObject.GPR_predict([x], estimate_err=True)
        return GPR_eval['y_gprErr'][0]

    if 'fitType' in res and res['fitType'] == 'GPR':
        gpr_predictObject = GPRPredictor(res)
        return gprErrEval
    else:
        raise Exception('getGPRErrorPrediction requires fitType=GPR.')


def getGPRFitAndErrorEvaluator(res):

    def gprEval(x):
        GPR_eval  = gpr_predictObject.GPR_predict([x], estimate_err=True)
        return GPR_eval['y'][0], GPR_eval['y_gprErr'][0]

    if 'fitType' in res and res['fitType'] == 'GPR':
        gpr_predictObject = GPRPredictor(res)
        return gprEval
    else:
        raise Exception('getGPRErrorPrediction requires fitType=GPR.')



import numpy as np
from surfinBH import surfinBH
import warnings

#=============================================================================
class Fit3dq8(surfinBH.SurFinBH):
    """ A class for the NRSur3dq8Remnant model presented in Varma et al.,
    arxiv:1809.09125. This model is referred to as surfinBH3dq8 in the paper.
    
    This model predicts the final mass mf, final
    spin chif and final kick velocity vf, for the remnants of nonprecessing
    binary black hole systems. The fits are done using Gaussian Process
    Regression (GPR) and also provide an error estimate along with the fit
    value.

    This model has been trained in the parameter space:
        q <= 8, |chiAz| <= 0.8, |chiBz| <= 0.8

    However, it extrapolates reasonably to:
        q <= 10, |chiAz| <= 1, |chiBz| <= 1

    =========================================================================
    Usage:

    import surfinBH

    # Load the fit
    fit = surfinBH.LoadFits('NRSur3dq8Remnant')

    We provide the following call methods:
        # remnant mass and 1-sigma error estimate
        mf, mf_err = fit.mf(q, chiA, chiB, **kwargs)

        # remnant spin and 1-sigma error estimate
        chif, chif_err = fit.chif(q, chiA, chiB, **kwargs)

        # remnant recoil kick and 1-sigma error estimate
        vf, vf_err = fit.vf(q, chiA, chiB, **kwargs)

        # All of these together
        mf, chif, vf, mf_err, chif_err, vf_err
            = fit.all(q, chiA, chiB, **kwargs)

    The arguments for each of these call methods are as follows:
    Arguments:
        q:      Mass ratio (q>=1)

        chiA:   Dimensionless spin of the larger BH (array of size 3).

        chiB:   Dimensionless spin of the smaller BH (array of size 3).
                This model allows only nonprecessing spins, so only the
                z-components of these arrays should be non-zero.

    Optional arguments:
        allow_extrap:
            If False, raises a warning when q > 8.1 or |chiA|,|chiB| > 0.81,
                and raises an error when q > 10.1 or |chiA|,|chiB| > 1.
            If True, allows extrapolation to any q and |chiA|,|chiB| <= 1.
                Use at your own risk.
            Default: False.

    The spin and kick vectors are defined in the coorbital frame at t=-100 M
    from the peak of the waveform. This frame is defined as:
    The z-axis is along the orbital angular momentum direction of the binary.
    The x-axis is along the line of separation from the smaller BH to
        the larger BH at this time.
    The y-axis completes the triad.
    We obtain this frame from the waveform as defined in arxiv:1705.07089.
    """

    #-------------------------------------------------------------------------
    def __init__(self, name):

        # Param limits beyond which to raise a warning
        soft_param_lims = {
            'q': 8.1,
            'chiAmag': 0.81,
            'chiBmag': 0.81,
                }

        # Param limits beyond which to raise an error
        hard_param_lims = {
            'q': 10.1,
            'chiAmag': 1,
            'chiBmag': 1,
                }
        super(Fit3dq8, self).__init__(name, soft_param_lims, hard_param_lims,
                aligned_spin_only=True)

    #-------------------------------------------------------------------------
    def _load_fits(self, h5file):
        """ Loads fits from h5file and returns a dictionary of fits. """
        fits = {}
        for key in ['mf', 'chifz', 'vfx', 'vfy']:
            fits[key] = self._load_scalar_fit(fit_key=key, h5file=h5file)
        return fits

    #-------------------------------------------------------------------------
    def _get_fit_params(self, x, fit_key):
        """ Transforms the input parameter to fit parameters for the 3dq8 model.
    That is, maps from [q, chiAz, chiBz] to [np.log(q), chiHat, chi_a]
    chiHat is defined in Eq.(3) of 1508.07253.
    chi_a = (chiAz - chiBz)/2.
        """
        q, chiAz, chiBz = x
        eta = q/(1.+q)**2
        chi_wtAvg = (q*chiAz+chiBz)/(1.+q)
        chiHat = (chi_wtAvg - 38.*eta/113.*(chiAz + chiBz))/(1. - 76.*eta/113.)
        chi_a = (chiAz - chiBz)/2.
        fit_params = [np.log(q), chiHat, chi_a]
        return fit_params

    #-------------------------------------------------------------------------
    def _eval_wrapper(self, fit_key, q, chiA, chiB, **kwargs):
        """ Evaluates the NRSur3dq8Remnant model.
        """
        chiA = np.array(chiA)
        chiB = np.array(chiB)

        # Warn/Exit if extrapolating
        allow_extrap = kwargs.pop('allow_extrap', False)
        self._check_param_limits(q, chiA, chiB, allow_extrap)

        self._check_unused_kwargs(kwargs)

        x = [q, chiA[2], chiB[2]]
        if fit_key == 'mf' or fit_key == 'all':
            mf, mf_err = self._evaluate_fits(x, 'mf')
            if fit_key == 'mf':
                return mf, mf_err
        if fit_key == 'chif' or fit_key == 'all':
            chifz, chifz_err = self._evaluate_fits(x, 'chifz')
            chif = np.array([0,0,chifz])
            chif_err = np.array([0,0,chifz_err])
            if fit_key == 'chif':
                return chif, chif_err
        if fit_key == 'vf' or fit_key == 'all':
            vfx, vfx_err = self._evaluate_fits(x, 'vfx')
            vfy, vfy_err = self._evaluate_fits(x, 'vfy')
            vf = np.array([vfx, vfy, 0])
            vf_err = np.array([vfx_err, vfy_err, 0])
            if fit_key == 'vf':
                return vf, vf_err
        if fit_key == 'all':
            return mf, chif, vf, mf_err, chif_err, vf_err



from . import surfinBH
__author__ = surfinBH.__author__
__email__ = surfinBH.__email__
__copyright__ = surfinBH.__copyright__
__license__ = surfinBH.__license__
__version__ = surfinBH.__version__
__doc__ = surfinBH.__doc__

from ._loadFits import LoadFits
from ._loadFits import fits_collection
from ._loadFits import DownloadData



import numpy as np

#-----------------------------------------------------------------------------
def multiplyQuats(q1, q2):
    """q1, q2 must be [scalar, x, y, z] but those may be arrays or scalars"""
    return np.array([
            q1[0]*q2[0] - q1[1]*q2[1] - q1[2]*q2[2] - q1[3]*q2[3],
            q1[2]*q2[3] - q2[2]*q1[3] + q1[0]*q2[1] + q2[0]*q1[1],
            q1[3]*q2[1] - q2[3]*q1[1] + q1[0]*q2[2] + q2[0]*q1[2],
            q1[1]*q2[2] - q2[1]*q1[2] + q1[0]*q2[3] + q2[0]*q1[3]])

#-----------------------------------------------------------------------------
def quatInv(q):
    """Returns QBar such that Q*QBar = 1"""
    qConj = -q
    qConj[0] = -qConj[0]
    normSqr = multiplyQuats(q, qConj)[0]
    return qConj/normSqr

#-----------------------------------------------------------------------------
def alignVec_quat(vec):
    """Returns a unit quaternion that will align vec with the z-axis"""
    alpha = np.arctan2(vec[1], vec[0])
    beta = np.arccos(vec[2])
    gamma = -alpha*vec[2]
    cb = np.cos(0.5*beta)
    sb = np.sin(0.5*beta)
    return np.array([cb*np.cos(0.5*(alpha + gamma)),
                     sb*np.sin(0.5*(gamma - alpha)),
                     sb*np.cos(0.5*(gamma - alpha)),
                     cb*np.sin(0.5*(alpha + gamma))])

#-----------------------------------------------------------------------------
def lHat_from_quat(quat):
    qInv = quatInv(quat)
    return multiplyQuats(quat, multiplyQuats(
                    np.array([0., 0., 0., 1.]), qInv))[1:]

#-----------------------------------------------------------------------------
def transformTimeDependentVector(quat, vec, inverse=0):
    """Given (for example) a minimal rotation frame quat, transforms
    vec from the minimal rotation frame to the inertial frame.
    With inverse=1, transforms from the inertial frame to the minimal
    rotation frame."""
    qInv = quatInv(quat)
    if inverse:
        return transformTimeDependentVector(qInv, vec, inverse=0)

    return multiplyQuats(quat, multiplyQuats(np.append(np.array([
            np.zeros(len(vec[0]))]), vec, 0), qInv))[1:]


#-----------------------------------------------------------------------------
def rotate_in_plane(chi, phase):
    """For transforming spins between the coprecessing and coorbital frames"""
    v = chi.T
    sp = np.sin(phase)
    cp = np.cos(phase)
    res = 1.*v
    res[0] = v[0]*cp + v[1]*sp
    res[1] = v[1]*cp - v[0]*sp
    return res.T

#-----------------------------------------------------------------------------
def transform_vector_coorb_to_inertial(vec_coorb, orbPhase, quat_copr):
    """Given a vector (of size 3) in coorbital frame, orbital phase in
    coprecessing frame and a minimal rotation frame quat, transforms
    the vector from the coorbital to the LAL inertial frame.
    """

    # Transform to coprecessing frame
    vec_copr = rotate_in_plane(vec_coorb, -orbPhase)

    # Transform to inertial frame (for the surrogate)
    vec = transformTimeDependentVector(np.array([quat_copr]).T,
        np.array([vec_copr]).T).T[0]

    return np.array(vec)


def transform_error_coorb_to_inertial(vec_coorb, vec_err_coorb, orbPhase,
        quat_copr):
    """ Transform error in a vector from the coorbital frame to the inertial
    frame. Generates distributions in the coorbital frame, transforms them
    to inertial frame and returns 1-simga widths in the inertial frame.
    """

    # for reproducibility
    np.random.seed(0)

    # Get distribution in coorbital frame
    dist_coorb = np.array([np.random.normal(m, s, 1000)
        for m,s in zip(vec_coorb, vec_err_coorb)]).T

    # Transform distribution to coprecessing frame
    dist_copr = rotate_in_plane(dist_coorb, -orbPhase)

    # Transform distribution to inertial frame (for the surrogate)
    dist_inertial = transformTimeDependentVector(
        np.array([quat_copr for _ in dist_copr]).T, dist_copr.T).T

    # Get 1sigma width in inertial frame
    vec_err_inertial = np.std(dist_inertial, axis=0)

    return vec_err_inertial



import numpy as np
from surfinBH import surfinBH
from surfinBH._lal_spin_evolution import evolve_pn_spins
import surfinBH._utils as utils
import warnings

#=============================================================================
class Fit7dq2(surfinBH.SurFinBH):
    """ A class for the surfinBH7dq2 model presented in Varma et al.,
    arxiv:1809.09125. This model predicts the final mass mf, final spin vector
    chif and final kick velocity vector vf, for the remnants of precessing
    binary black hole systems.  The fits are done using Gaussian Process
    Regression (GPR) and also provide an error estimate along with the fit
    value.

    This model has been trained in the parameter space:
        q <= 2, |chiA| <= 0.8, |chiB| <= 0.8

    However, it extrapolates reasonably to:
        q <= 4, |chiA| <= 1, |chiB| <= 1

    =========================================================================
    Usage:

    import surfinBH

    # Load the fit
    fit = surfinBH.LoadFits('surfinBH7dq2')

    We provide the following call methods:
        # remnant mass and 1-sigma error estimate
        mf, mf_err = fit.mf(q, chiA, chiB, **kwargs)

        # remnant spin and 1-sigma error estimate
        chif, chif_err = fit.chif(q, chiA, chiB, **kwargs)

        # remnant recoil kick and 1-sigma error estimate
        vf, vf_err = fit.vf(q, chiA, chiB, **kwargs)

        # All of these together
        mf, chif, vf, mf_err, chif_err, vf_err
            = fit.all(q, chiA, chiB, **kwargs)

    The arguments for each of these call methods are as follows:
    Arguments:
        q:      Mass ratio (q>=1)

        chiA:   Dimensionless spin of the larger BH (array of size 3)

        chiB:   Dimensionless spin of the smaller BH (array of size 3)

            By default, the spins are assumed to be the component spins at
            t=-100 M from the peak of the waveform, and in the coorbital frame,
            defined as:
            The z-axis is along the orbital angular momentum at t=-100M.
            The x-axis is along the line of separation from the smaller BH to
                the larger BH at this time.
            The y-axis completes the triad.
            We obtain this frame from the waveform as defined in
            arxiv:1705.07089.
            The returned spin and kick vectors are also in the same frame.

            If 'omega0' is given, instead the spins are assumed to be the
            component spins when the PN orbital frequency = omega0. The
            spins are assumed to be in the inertial frame, defined as:
            The z-axis is along the Newtonian orbital angular momentum when the
                PN orbital frequency = omega0.
            The x-axis is along the line of separation from the smaller BH to
                the larger BH at this frequency.
            The y-axis completes the triad.
            We obtain this frame from PN.
            Given the spins at omega0, we evolve the spins using PN until
            the orbital frequency = omega_switch, then we further evolve the
            spins using the NRSur7dq2 model (arxiv:1705.07089) until t=-100M
            from the peak of the waveform. Then we evaluate the fits using the
            spins at t=-100 M. Finally, we transform the remnant spin and kick
            vectors back to the inertial frame defined above.

    Optional arguments:
        allow_extrap:
            If False, raises a warning when q > 2.1 or |chiA|,|chiB| > 0.81,
                and raises an error when q > 4.1 or |chiA|,|chiB| > 1.
            If True, allows extrapolation to any q and |chiA|,|chiB| <= 1.
                Use at your own risk.
            Default: False.

        omega0:
            Initial dimensionless orbital frequency in units of 1/M, where M is
            the total mass. If this is given, the spins in x are assumed to be
            the component spins at this orbital frequency, and in the inertial
            frame as defined above. The returned remnant spin and kick vectors
            are also in the same frame.
            Default: None.

        PN_approximant:
            Approximant used to do the PN spin evolution. Choose from
            'SpinTaylorT4', 'SpinTaylorT1' or 'SpinTaylorT5'.
            Default: 'SpinTaylorT4'.

        PN_dt:
            Dimensionless time step size in units of M (total mass), used for
            the PN evolution. You may need to increase this if omega0 is very
            low.
            Default: 0.1

        PN_spin_order:
            Twice the PN order of spin effects. E.g., use 7 for 3.5PN.
            Default: 7

        PN_phase_order:
            Twice the PN order in phase. E.g., use 7 for 3.5PN.
            Default: 7

        omega_switch:
            Dimensionless orbital frequency at which to switch from PN to
            NRSur7dq2 model. You may need to increase this if the NRSur7dq2
            model raises an exception like:
            "Got omega_ref = 0.0180 < 0.0184 = omega_0, too small!"
            Default: 0.018
    """

    #-------------------------------------------------------------------------
    def __init__(self, name, load_nrsur=False):

        # Param limits beyond which to raise a warning
        soft_param_lims = {
            'q': 2.1,
            'chiAmag': 0.81,
            'chiBmag': 0.81,
                }

        # Param limits beyond which to raise an error
        hard_param_lims = {
            'q': 4.1,
            'chiAmag': 1,
            'chiBmag': 1,
                }

        super(Fit7dq2, self).__init__(name, soft_param_lims, hard_param_lims)
        self.nrsur = None

    #-------------------------------------------------------------------------
    def _load_NRSur7dq2(self):
        import NRSur7dq2
        self.nrsur = NRSur7dq2.NRSurrogate7dq2()
        print('Loaded NRSur7dq2 waveform model')

    #-------------------------------------------------------------------------
    def _load_fits(self, h5file):
        """ Loads fits from h5file and returns a dictionary of fits. """
        fits = {}
        for key in ['mf']:
            fits[key] = self._load_scalar_fit(fit_key=key, h5file=h5file)
        for key in ['chif', 'vf']:
            fits[key] = self._load_vector_fit(key, h5file)
        return fits

    #-------------------------------------------------------------------------
    def _extra_regression_kwargs(self):
        """ List of additional kwargs to use in regression tests.
        """

        # larger than default sometimes needed when extrapolating
        omega_switch_test = 0.019

        extra_args = []
        extra_args.append({
            'omega0': 5e-3,
            'PN_approximant': 'SpinTaylorT4',
            'PN_dt': 0.1,
            'PN_spin_order': 7,
            'PN_phase_order': 7,
            'omega_switch': omega_switch_test,
            })


        extra_args.append({
            'omega0': 6e-3,
            'PN_approximant': 'SpinTaylorT1',
            'PN_dt': 0.5,
            'PN_spin_order': 5,
            'PN_phase_order': 7,
            'omega_switch': omega_switch_test,
            })

        extra_args.append({
            'omega0': 7e-3,
            'PN_approximant': 'SpinTaylorT5',
            'PN_dt': 1,
            'PN_spin_order': 7,
            'PN_phase_order': 5,
            'omega_switch': omega_switch_test,
            })

        # These should be pure NRSur7dq2
        extra_args.append({'omega0': 3e-2})
        extra_args.append({'omega0': 5e-2})

        return extra_args

    #-------------------------------------------------------------------------
    def _get_fit_params(self, x, fit_key):
        """ Transforms the input parameter to fit parameters for the 7dq2 model.
    That is, maps from
    x = [q, chiAx, chiAy, chiAz, chiBx, chiBy, chiBz]
    fit_params = [np.log(q), chiAx, chiAy, chiHat, chiBx, chiBy, chi_a]

    chiHat is defined in Eq.(3) of 1508.07253, but with chiAz and chiBz instead
    of chiA and chiB.
    chi_a = (chiAz - chiBz)/2.
        """
        q, chiAz, chiBz = x[0], x[3], x[6]
        eta = q/(1.+q)**2
        chi_wtAvg = (q*chiAz+chiBz)/(1.+q)
        chiHat = (chi_wtAvg - 38.*eta/113.*(chiAz + chiBz))/(1. - 76.*eta/113.)
        chi_a = (chiAz - chiBz)/2.

        fit_params = [np.log(q), x[1], x[2], chiHat, x[4], x[5], chi_a]
        return fit_params

    #-------------------------------------------------------------------------
    def _evolve_spins(self, q, chiA0, chiB0, omega0, PN_approximant,
            PN_dt, PN_spin0, PN_phase0, omega0_nrsur):
        """ Evolves spins of the component BHs from an initial orbital
        frequency = omega0 until t=-100 M from the peak of the waveform.
        If omega0 < omega0_nrsur, use PN to evolve the spins until
        orbital frequency = omega0. Then evolves further with the NRSur7dq2
        waveform model until t=-100M from the peak.

        Assumes chiA0 and chiB0 are defined in the inertial frame defined
        at orbital frequency = omega0 as:
            The z-axis is along the Newtonian orbital angular momentum when the
                PN orbital frequency = omega0.
            The x-axis is along the line of separation from the smaller BH to
                the larger BH at this frequency.
            The y-axis completes the triad.

        Returns spins in the coorbital frame at t=-100M, as well as the
        coprecessing frame quaternion and orbital phase in the coprecessing
        frame at this time.
        """

        if omega0 < omega0_nrsur:
            # If omega0 is below the NRSur7dq2 start frequency, we use PN
            # to evolve the spins until orbital frequency = omega0_nrsur.

            # Note that we update omega0_nrsur here with the PN
            # frequency that was closest to the input omega0_nrsur.
            chiA0_nrsur_copr, chiB0_nrsur_copr, quat0_nrsur_copr, \
                phi0_nrsur, omega0_nrsur \
                = evolve_pn_spins(q, chiA0, chiB0, omega0,
                    omega0_nrsur, approximant=PN_approximant,
                    dt=PN_dt, spinO=PN_spin0,
                    phaseO=PN_phase0)
        else:
            # If omega0>= omega0_nrsur, we evolve spins directly with NRSur7dq2
            # waveform model. We set the coprecessing frame quaternion to
            # identity and orbital phase to 0 at omega=omega0, hence the
            # coprecessing frame is the same as the inertial frame here.

            # Note that we update omega0_nrsur here and set it to omega0
            chiA0_nrsur_copr, chiB0_nrsur_copr, quat0_nrsur_copr, \
                phi0_nrsur, omega0_nrsur \
                = chiA0, chiB0, [1,0,0,0], 0, omega0

        # Load NRSur7dq2 if needed
        if self.nrsur is None:
            self._load_NRSur7dq2()

        # evaluate NRSur7dq2 dynamics
        # We set allow_extrapolation=True always since we test param limits
        # independently
        quat, orbphase, chiA_copr, chiB_copr = self.nrsur.get_dynamics(q,
            chiA0_nrsur_copr, chiB0_nrsur_copr, init_quat=quat0_nrsur_copr,
            init_phase=phi0_nrsur, omega_ref=omega0_nrsur,
            allow_extrapolation=True)

        # get data at time node where remnant fits are done
        fitnode_time = -100
        nodeIdx = np.argmin(np.abs(self.nrsur.tds - fitnode_time))
        quat_fitnode = quat.T[nodeIdx]
        orbphase_fitnode = orbphase[nodeIdx]

        # get coorbital frame spins at the time node
        chiA_coorb_fitnode = utils.rotate_in_plane(chiA_copr[nodeIdx],
                orbphase_fitnode)
        chiB_coorb_fitnode = utils.rotate_in_plane(chiB_copr[nodeIdx],
                orbphase_fitnode)

        return chiA_coorb_fitnode, chiB_coorb_fitnode, quat_fitnode, \
            orbphase_fitnode

    #-------------------------------------------------------------------------
    def _eval_wrapper(self, fit_key, q, chiA, chiB, **kwargs):
        """Evaluates the surfinBH7dq2 model.
        """
        chiA = np.array(chiA)
        chiB = np.array(chiB)

        # Warn/Exit if extrapolating
        allow_extrap = kwargs.pop('allow_extrap', False)
        self._check_param_limits(q, chiA, chiB, allow_extrap)

        omega0 = kwargs.pop('omega0', None)
        PN_approximant = kwargs.pop('PN_approximant', 'SpinTaylorT4')
        PN_dt = kwargs.pop('PN_dt', 0.1)
        PN_spin_order = kwargs.pop('PN_spin_order', 7)
        PN_phase_order = kwargs.pop('PN_phase_order', 7)
        omega_switch = kwargs.pop('omega_switch', 0.018)

        self._check_unused_kwargs(kwargs)

        if omega0 is None:
            # If omega0 is given, assume chiA, chiB are the coorbital frame
            # spins at t=-100 M.
            x = np.concatenate(([q], chiA, chiB))
        else:
            # If omega0 is given, evolve the spins from omega0
            # to t = -100 M from the peak.
            chiA_coorb_fitnode, chiB_coorb_fitnode, quat_fitnode, \
                orbphase_fitnode \
                = self._evolve_spins(q, chiA, chiB, omega0,
                    PN_approximant, PN_dt, PN_spin_order,
                    PN_phase_order, omega_switch)
            # x should contain coorbital frame spins at t=-100M
            x = np.concatenate(([q], chiA_coorb_fitnode, chiB_coorb_fitnode))


        def eval_vector_fit(x, fit_key):
            res = self._evaluate_fits(x, fit_key)
            fit_val = res.T[0]
            fit_err = res.T[1]
            if omega0 is not None:
                # If spins were given in inertial frame at omega0,
                # transform vectors and errors back to the same frame.
                fit_val = utils.transform_vector_coorb_to_inertial(fit_val,
                    orbphase_fitnode, quat_fitnode)
                fit_err = utils.transform_error_coorb_to_inertial(fit_val,
                    fit_err, orbphase_fitnode, quat_fitnode)
            return fit_val, fit_err


        if fit_key == 'mf' or fit_key == 'all':
            mf, mf_err = self._evaluate_fits(x, 'mf')
            if fit_key == 'mf':
                return mf, mf_err

        if fit_key == 'chif' or fit_key == 'all':
            chif, chif_err = eval_vector_fit(x, 'chif')
            if fit_key == 'chif':
                return chif, chif_err

        if fit_key == 'vf' or fit_key == 'all':
            vf, vf_err = eval_vector_fit(x, 'vf')
            if fit_key == 'vf':
                return vf, vf_err

        if fit_key == 'all':
            return mf, chif, vf, mf_err, chif_err, vf_err



"""surfinBH
========

Surrogate final black hole properties for mergers of binary black holes.
See https://pypi.org/project/surfinBH/ for more details.
"""
__copyright__ = "Copyright (C) 2018 Vijay Varma"
__email__ = "vvarma@caltech.edu"
__status__ = "testing"
__author__ = "Vijay Varma"
__version__ = "1.1.3"
__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import numpy as np
import os, sys
import h5py
import warnings

from . import _eval_pysur
from ._dataPath import DataPath


#=============================================================================
class SurFinBH(object):
    """
Class to load and evaluate surrogate fits for final BH properties.
Each derived class should do the following:

    1. define _load_fits(self, h5file)
    2. define _get_fit_params(self, x, fit_key)
    3. define _eval_wrapper(self, fit_key, x, **kwargs)
    4. define soft_param_lims and hard_param_lims.
    5. define _extra_regression_kwargs, to test any additional kwargs used in
          the _eval_wrapper method.

See _fit_evaluators.fit_7dq2.py for an example.
    """

    #-------------------------------------------------------------------------
    def __init__(self, name, soft_param_lims, hard_param_lims,
            aligned_spin_only=False):
        """
        name: Name of the fit excluding the surfinBH prefix. Ex: 7dq2.
        soft_param_lims: param limits beyond which to raise a warning.
        hard_param_lims: param limits beyond which to raise an error.
        aligned_spin_only: raise an error if given precessing spins.
        See _fit_evaluators.fit_7dq2.py for an example.
        """
        self.name = name
        self.soft_param_lims = soft_param_lims
        self.hard_param_lims = hard_param_lims
        self.aligned_spin_only = aligned_spin_only

        h5file = h5py.File('%s/fit_%s.h5'%(DataPath(), name), 'r')
        self.fits = self._load_fits(h5file)
        h5file.close()

    #-------------------------------------------------------------------------
    def _read_dict(self, f):
        """ Converts h5 groups to dictionaries
        """
        d = {}
        for k, item in f.items():
            if type(item) == h5py._hl.dataset.Dataset:
                v = item[()]
                if type(v) == np.string_:
                    v = str(v)
                if type(v) == str and v == "NONE":
                    d[k] = None
                elif type(v) == str and v == "EMPTYARR":
                    d[k] = np.array([])
                elif isinstance(v, bytes):
                    d[k] = v.decode('utf-8')
                else:
                    d[k] = v

            elif k[:5] == "DICT_":
                d[k[5:]] = self._read_dict(item)
            elif k[:5] == "LIST_":
                tmpD = self._read_dict(item)
                d[k[5:]] = [tmpD[str(i)] for i in range(len(tmpD))]
        return d

    #-------------------------------------------------------------------------
    def _load_scalar_fit(self, fit_key=None, h5file=None, fit_data=None):
        """ Loads a single fit
        """
        if (fit_key is None) ^ (h5file is None):
            raise ValueError("Either specify both fit_key and h5file, or"
                " neither")

        if not ((fit_key is None) ^ (fit_data is None)):
            raise ValueError("Specify exactly one of fit_key and fit_data.")

        if fit_data is None:
            fit_data = self._read_dict(h5file[fit_key])

        if 'fitType' in fit_data.keys() and fit_data['fitType'] == 'GPR':
            fit = _eval_pysur.evaluate_fit.getGPRFitAndErrorEvaluator(fit_data)
        else:
            fit = _eval_pysur.evaluate_fit.getFitEvaluator(fit_data)

        return fit

    #-------------------------------------------------------------------------
    def _load_vector_fit(self, fit_key, h5file):
        """ Loads a vector of fits
        """
        vector_fit = []
        for i in range(len(h5file[fit_key].keys())):
            fit_data = self._read_dict(h5file[fit_key]['comp_%d'%i])
            vector_fit.append(self._load_scalar_fit(fit_data=fit_data))
        return vector_fit

    #-------------------------------------------------------------------------
    def _evaluate_fits(self, x, fit_key):
        """ Evaluates a particular fit by passing fit_key to self.fits.
            Assumes self._get_fit_params() has been overriden.
        """
        fit = self.fits[fit_key]
        fit_params = self._get_fit_params(np.copy(x), fit_key)
        if type(fit) == list:
            res = []
            for i in range(len(fit)):
                res.append(fit[i](fit_params))
            return np.array(res)
        else:
            return fit(fit_params)

    #-------------------------------------------------------------------------
    def _check_unused_kwargs(self, kwargs):
        """ Call this at the end of call module to check if all the kwargs have
        been used. Assumes kwargs were extracted using pop.
        """
        if len(kwargs.keys()) != 0:
            unused = ""
            for k in kwargs.keys():
                unused += "'%s', "%k
            if unused[-2:] == ", ":     # get rid of trailing comma
                unused = unused[:-2]
            raise Exception('Unused keys in kwargs: %s'%unused)

    #-------------------------------------------------------------------------
    def _check_param_limits(self, q, chiA, chiB, allow_extrap):
        """ Checks that params are within allowed range of paramters.
        Raises a warning if outside self.soft_param_lims limits and
        raises an error if outside self.hard_param_lims.
        If allow_extrap=True, skips these checks.
        """

        if q < 1:
            raise ValueError('Mass ratio should be >= 1.')

        chiAmag = np.sqrt(np.sum(chiA**2))
        chiBmag = np.sqrt(np.sum(chiB**2))
        if chiAmag > 1 + 1e-14:
            raise ValueError('Spin magnitude of BhA > 1.')
        if chiBmag > 1 + 1e-14:
            raise ValueError('Spin magnitude of BhB > 1.')

        if self.aligned_spin_only:
            if np.sqrt(np.sum(chiA[:2]**2)) > 1e-14:
                raise ValueError('The x & y components of chiA should be zero.')
            if np.sqrt(np.sum(chiB[:2]**2)) > 1e-14:
                raise ValueError('The x & y components of chiB should be zero.')

        # Do not check param limits if allow_extrap=True
        if allow_extrap:
            return

        if q > self.hard_param_lims['q']+ 1e-14:
            raise ValueError('Mass ratio outside allowed range.')
        elif q > self.soft_param_lims['q']:
            warnings.warn('Mass ratio outside training range.')

        if chiAmag > self.hard_param_lims['chiAmag']+ 1e-14:
            raise ValueError('Spin magnitude of BhA outside allowed range.')
        elif chiAmag > self.soft_param_lims['chiAmag']:
            warnings.warn('Spin magnitude of BhA outside training range.')

        if chiBmag > self.hard_param_lims['chiBmag']+ 1e-14:
            raise ValueError('Spin magnitude of BhB outside allowed range.')
        elif chiBmag > self.soft_param_lims['chiBmag']:
            warnings.warn('Spin magnitude of BhB outside training range.')

    #-------------------------------------------------------------------------
    def _generate_random_params_for_tests(self):
        """ Generate random parameters to use in tests.
        """
        # Generate params randomly within allowed values
        q = np.random.uniform(1, self.hard_param_lims['q'])
        chiAmag= np.random.uniform(0, self.hard_param_lims['chiAmag'])
        chiBmag= np.random.uniform(0, self.hard_param_lims['chiBmag'])
        if self.aligned_spin_only:
            chiAph, chiBph = 0, 0
            chiAth, chiBth = np.random.choice([0, np.pi]), \
                np.random.choice([0, np.pi])
        else:
            chiAth = np.arccos(np.random.uniform(-1., 1.))
            chiBth = np.arccos(np.random.uniform(-1., 1.))
            chiAph = np.random.uniform(0, 2*np.pi)
            chiBph = np.random.uniform(0, 2*np.pi)

        chiA = [chiAmag*np.sin(chiAth)*np.cos(chiAph),
                chiAmag*np.sin(chiAth)*np.sin(chiAph),
                chiAmag*np.cos(chiAth)]

        chiB = [chiBmag*np.sin(chiBth)*np.cos(chiBph),
                chiBmag*np.sin(chiBth)*np.sin(chiBph),
                chiBmag*np.cos(chiBth)]

        return q, chiA, chiB


    #-------------------------------------------------------------------------
    #----------------------  Override these  ---------------------------------
    #-------------------------------------------------------------------------

    #-------------------------------------------------------------------------
    def _load_fits(self, h5file):
        """ Loads fits from h5file and returns a dictionary of fits. """
        raise NotImplementedError("Please override me.")
        return fits


    #-------------------------------------------------------------------------
    def _extra_regression_kwargs(self):
        """ Add additional kwargs for regression tests. If not overriden,
            this will be empty. See _fit_evaluators.fit_7dq2.py for an example.
        """
        return []

    #-------------------------------------------------------------------------
    def _get_fit_params(self, x, fit_key):
        """ Maps from input params x to the fit_params used to evaluate the
            fit.
        """
        raise NotImplementedError("Please override me.")
        return fit_params

    #-------------------------------------------------------------------------
    def _eval_wrapper(self, fit_key, q, chiA, chiB, **kwargs):
        """ Evaluates a particular fit. This varies for each surrogate.
            Allowed values for fit_key are 'mf', 'chif' and 'vf' and 'all'.
            chiA and chiB should have size 3.

            Each derived class should have its own _eval_wrapper function but
            call self._check_param_limits() first to do some sanity checks.
            See _fit_evaluators.fit_7dq2.py for an example.
        """
        raise NotImplementedError("Please override me.")


    #-------------------------------------------------------------------------
    #----------------------   Call methods   ---------------------------------
    #-------------------------------------------------------------------------

    def mf(self, *args, **kwargs):
        """ Evaluates fit and 1-sigma error estimate for remnant mass.
        Returns:
            mf, mf_err_est
        """
        return self._eval_wrapper('mf', *args, **kwargs)

    def chif(self, *args, **kwargs):
        """ Evaluates fit and 1-sigma error estimate for remnant spin.
        Returns:
            chif, chif_err_est

        chif and chif_err_est are arrays of size 3.
        """
        return self._eval_wrapper('chif', *args, **kwargs)

    def vf(self, *args, **kwargs):
        """ Evaluates fit and 1-sigma error estimate for remnant kick velocity.
        Returns:
            vf, vf_err_est

        vf and vf_err_est are arrays of size 3.
        """
        return self._eval_wrapper('vf', *args, **kwargs)

    def all(self, *args, **kwargs):
        """ Evaluates fit and 1-sigma error estimate for remnant mass, spin
        and kick velocity.
        Returns:
            mf, chif, vf, mf_err_est, chif_err_est, vf_err_est

        chif, vf, chif_err_est and vf_err_est are arrays of size 3.
        """
        return self._eval_wrapper('all', *args, **kwargs)




import numpy as np
import sys
from scipy.interpolate import InterpolatedUnivariateSpline as spline
from surfinBH import surfinBH
from surfinBH._lal_spin_evolution import lal_spin_evloution_wrapper
import surfinBH._utils as utils
import warnings

#=============================================================================
class Fit7dq4(surfinBH.SurFinBH):
    """ A class for the NRSur7dq4Remnant model presented in Varma et al.,
    arxiv:1905.09300, hereafter referred to as THE PAPER.

    This model predicts the final mass mf, final spin vector
    chif and final kick velocity vector vf, for the remnants of precessing
    binary black hole systems.  The fits are done using Gaussian Process
    Regression (GPR) and also provide an error estimate along with the fit
    value.

    This model has been trained in the parameter space:
        q <= 4, |chiA| <= 0.8, |chiB| <= 0.8

    However, it extrapolates reasonably to:
        q <= 6, |chiA| <= 1, |chiB| <= 1

    =========================================================================
    Usage:

    import surfinBH

    # Load the fit
    fit = surfinBH.LoadFits('NRSur7dq4Remnant')

    We provide the following call methods:
        # remnant mass and 1-sigma error estimate
        mf, mf_err = fit.mf(q, chiA, chiB, **kwargs)

        # remnant spin and 1-sigma error estimate
        chif, chif_err = fit.chif(q, chiA, chiB, **kwargs)

        # remnant recoil kick and 1-sigma error estimate (units of c)
        vf, vf_err = fit.vf(q, chiA, chiB, **kwargs)

        # All of these together
        mf, chif, vf, mf_err, chif_err, vf_err
            = fit.all(q, chiA, chiB, **kwargs)

    The arguments for each of these call methods are as follows:
    Arguments:
        q:      Mass ratio (q = mA/mB >= 1)

        chiA:  Dimensionless spin vector of the heavier black hole at
                reference epoch.
        chiB:  Dimensionless spin vector of the lighter black hole at
                reference epoch.

                This follows the same convention as LAL, where the spin
                components are defined as:
                \chi_z = \chi \cdot \hat{L}, where L is the orbital angular
                    momentum vector at the epoch.
                \chi_x = \chi \cdot \hat{n}, where n = body2 -> body1 is the
                    separation vector at the epoch. body1 is the heavier body.
                \chi_y = \chi \cdot \hat{L \cross n}.
                These spin components are frame-independent as they are defined
                using vector inner products. This is equivalent to specifying
                the spins in the coorbital frame at the reference epoch. See
                THE PAPER for a definition of the coorbital frame.


    Optional arguments:

        omega0: Orbital frequency used to set the reference epoch.
                Default: None.

                If omega0 is None, the reference epoch is assumed to be at
                t=-100 M from the peak of the waveform, see THE PAPER for
                definition of the peak.

                If 'omega0' is given, the reference epoch is take to be the
                time at which the orbital frequency in the coprecessing frame
                equals omega0. omega0 should be in dimensionless units of
                rad/M, where M is the total mass.

                See THE PAPER for how the orbital frequency is
                computed as well as the definition of the coprecessing frame.

        allow_extrap:
            If False, raises a warning when q > 4.1 or |chiA|,|chiB| > 0.81,
                and raises an error when q > 6.1 or |chiA|,|chiB| > 1.
            If True, allows extrapolation to any q and |chiA|,|chiB| <= 1.
                Use at your own risk.
            Default: False.

    Optional PN evolution arguments:

        If the omega0 option is used, the spins need to be evolved from omega0
        until t=-100M, where the fits will be evaluated. For the late inspiral
        part, we use the internal spin evolution of NRSur7dq4 (also described
        in THE PAPER), which is very accurate. However, this surrogate is not
        long enough for small values of omega0 as it only has data starting at
        t=-4300M. Therefore, whenever the input omega0 is smaller than
        omega_switch_IG (defined below), we use PN evolution to go from omega0
        to about t=-4300M, beyond which we use NRSur7dq4 for spin evolution.

        PN_approximant:
            Approximant used to do the PN spin evolution. Choose from
            'SpinTaylorT4', 'SpinTaylorT1' or 'SpinTaylorT5'.
            Default: 'SpinTaylorT4'.

        PN_dt:
            Dimensionless time step size in units of M, used for the PN
            evolution. You may need to increase this if omega0 is very low.
            Default: 0.1

        PN_spin_order:
            Twice the PN order of spin effects. E.g., use 7 for 3.5PN.
            Default: 7

        PN_phase_order:
            Twice the PN order in phase. E.g., use 7 for 3.5PN.
            Default: 7

        t_sur_switch:
            The dimensionless time (from the peak) at which we switch from PN
            to the surrogate. Should be something larger than -4300.
            Default: -4000.

        omega_switch_IG:
            Initial guess for dimensionless orbital frequency, using which the
            switch will be made from PN to NRSur7dq4. This should be large
            enough to work for generic parts of the surrogate parameter space.
            You may need to increase this if the NRSur7dq4 model raises an
            exception like: "Got omega_ref=0.03 < 0.031=omega_0, too small!"
            Default: 0.03

            How t_sur_switch and omega_switch_IG work: The PN data is first
            generated starting at omega0, then the PN spins at omega_switch_IG
            are used to generate the NRSur7dq4 dynamics. NRSur7dq4 integrate
            the dynamics both forwards and backwards, so it will have omega and
            spins as a time series starting from -4300M. This is used to pick
            the omega0_sur and spins at t_sur_switch. Then the surrogate
            is reevaluated using omega0_sur and spins at t_sur_switch, thus
            ensuring that the switch always happens at t_sur_switch, even if
            omega_switch_IG corresponds to a later time.

    Inertial frame for returned values:

        The returned chif/vf are in the LAL inertial frame defined as follows:
            The +ve z-axis is along the orbital angular momentum at the
            reference epoch. The separation vector from the lighter BH to the
            heavier BH at the reference epoch is along the +ve x-axis. The
            y-axis completes the right-handed triad.

            Note that the default reference epoch corresponds to t=-100M, but
            if omega0 is given the reference epoch is taken to be the time at
            which the orbital frequency in the coprecessing frame is equal to
            omega0. This agrees with the LAL convention. See LIGO DCC document
            T1800226 for the LAL frame diagram.
    """

    #-------------------------------------------------------------------------
    def __init__(self, name, load_nrsur=False):

        # Param limits beyond which to raise a warning
        soft_param_lims = {
            'q': 4.1,
            'chiAmag': 0.81,
            'chiBmag': 0.81,
                }

        # Param limits beyond which to raise an error
        hard_param_lims = {
            'q': 6.1,
            'chiAmag': 1,
            'chiBmag': 1,
                }

        super(Fit7dq4, self).__init__(name, soft_param_lims, hard_param_lims)
        self.nrsur = None
        self.fitnode_time = -100      # Time at which the fits are constructed

    #-------------------------------------------------------------------------
    def _load_NRSur7dq4(self):
        import gwsurrogate
        from gwsurrogate.new.precessing_surrogate import splinterp_many
        self.nrsur = gwsurrogate.LoadSurrogate('NRSur7dq4')
        self.splinterp_many = splinterp_many

    #-------------------------------------------------------------------------
    def _load_fits(self, h5file):
        """ Loads fits from h5file and returns a dictionary of fits. """
        fits = {}
        for key in ['mf']:
            fits[key] = self._load_scalar_fit(fit_key=key, h5file=h5file)
        for key in ['chif', 'vf']:
            fits[key] = self._load_vector_fit(key, h5file)
        return fits

    #-------------------------------------------------------------------------
    def _extra_regression_kwargs(self):
        """ List of additional kwargs to use in regression tests.
        """

        extra_args = []
        extra_args.append({
            'omega0': 5e-3,
            'PN_approximant': 'SpinTaylorT4',
            'PN_dt': 0.1,
            'PN_spin_order': 7,
            'PN_phase_order': 7,
            })


        extra_args.append({
            'omega0': 6e-3,
            'PN_approximant': 'SpinTaylorT1',
            'PN_dt': 0.5,
            'PN_spin_order': 5,
            'PN_phase_order': 7,
            })

        extra_args.append({
            'omega0': 7e-3,
            'PN_approximant': 'SpinTaylorT5',
            'PN_dt': 1,
            'PN_spin_order': 7,
            'PN_phase_order': 5,
            })

        # These should be pure NRSur7dq4
        extra_args.append({'omega0': 3e-2})
        extra_args.append({'omega0': 5e-2})

        return extra_args

    #-------------------------------------------------------------------------
    def _get_fit_params(self, x, fit_key):
        """ Transforms the input parameter to fit parameters for the 7dq4 model.
    That is, maps from
    x = [q, chiAx, chiAy, chiAz, chiBx, chiBy, chiBz]
    fit_params = [np.log(q), chiAx, chiAy, chiHat, chiBx, chiBy, chi_a]

    chiHat is defined in Eq.(3) of 1508.07253, but with chiAz and chiBz instead
    of chiA and chiB.
    chi_a = (chiAz - chiBz)/2.
        """
        q, chiAz, chiBz = x[0], x[3], x[6]
        eta = q/(1.+q)**2
        chi_wtAvg = (q*chiAz+chiBz)/(1.+q)
        chiHat = (chi_wtAvg - 38.*eta/113.*(chiAz + chiBz))/(1. - 76.*eta/113.)
        chi_a = (chiAz - chiBz)/2.

        fit_params = [np.log(q), x[1], x[2], chiHat, x[4], x[5], chi_a]
        return fit_params

    def _get_coorbital_frame_spins_at_idx(self, chiA, chiB, omega, lNhat, phi, \
            idx):
        """ Computes PN spins and dynamics at a given idx.

        Inputs:
            chiA:       Dimless spin evolution of BhA in inertial frame.
            chiB:       Dimless spin evolution of BhB in inertial frame.
            omega:      Orbital frequency evolution in dimless units.
            lNhat:      Orbital angular momentum direction evolution.
            phi:        Orbital phase evolution.
            idx:        Index for output.

        Outputs (all are time series):
            chiA_at_idx_coorb:   Spin of BhA at idx, in coorbital frame.
            chiB_at_idx_coorb:   Spin of BhB at idx, in coorbital frame.
            quat_copr_at_idx:   Coprecessing frame quaternion at idx.
            phi_at_idx:         Orbital phase in the coprecessing frame at idx.
            omega_at_idx        Dimensionless orbital frequency at idx.

        The inertial frame is assumed to be aligned to the coorbital frame at
        the first index.
        """

        # Compute omega, inertial spins, angular momentum direction and orbital
        # phase at idx
        omega_at_idx = omega[idx]
        chiA_at_idx = chiA[idx]
        chiB_at_idx = chiB[idx]
        lNhat_at_idx = lNhat[idx]
        phi_at_idx = phi[idx]

        # Align the z-direction along orbital angular momentum direction
        # at idx. This moves us in to the coprecessing frame.
        quat_copr_at_idx = utils.alignVec_quat(lNhat_at_idx)
        chiA_at_idx_copr = utils.transformTimeDependentVector(
                np.array([quat_copr_at_idx]).T,
                np.array([chiA_at_idx]).T, inverse=1).T[0]
        chiB_at_idx_copr = utils.transformTimeDependentVector(
                np.array([quat_copr_at_idx]).T,
                np.array([chiB_at_idx]).T, inverse=1).T[0]

        # get coorbital frame spins at idx
        chiA_at_idx_coorb = utils.rotate_in_plane(chiA_at_idx_copr, phi_at_idx)
        chiB_at_idx_coorb = utils.rotate_in_plane(chiB_at_idx_copr, phi_at_idx)

        return chiA_at_idx_coorb, chiB_at_idx_coorb, quat_copr_at_idx, \
            phi_at_idx, omega_at_idx

    def _get_PN_spins_at_surrogate_start(self, PN_approximant, q, omega0, \
            chiA0, chiB0, PN_dt, PN_spin_order, PN_phase_order, \
            omega_switch_IG, t_sur_switch):
        """ Computes PN spins and frame dynamics at a time close to the start
            of the surrogate waveform model.

            Generates PN spins and frame quantities using spins at omega0.
            Then uses the PN spins at omega_switch_IG to generate the surrogate
            dynamics.
            Then gets the surrogate orbital frequency at t_sur_switch, let's
            call this omega_init_sur.
            Then use the PN spins at omega_init_sur to regenerate the surrogate
            dynamics.
        """

        # Get PN spin evolution starting at omega0
        omega_PN, phi_PN, chiA_PN, chiB_PN, lNhat_PN, e1_PN \
                = lal_spin_evloution_wrapper(PN_approximant, q, omega0, \
                chiA0, chiB0, PN_dt, PN_spin_order, PN_phase_order)

        # Get PN coorbital frame spins and frame dynamics at
        # omega_PN=omega_switch_IG
        idx = np.argmin(np.abs(omega_PN - omega_switch_IG))
        chiA_PN_at_idx_coorb, chiB_PN_at_idx_coorb, quat_PN_copr_at_idx, \
            phi_PN_at_idx, omega_PN_at_idx \
            = self._get_coorbital_frame_spins_at_idx(chiA_PN, chiB_PN, \
            omega_PN, lNhat_PN, phi_PN, idx)

        # Now evaluate the surrogate dynamics (both forwards and backwards)
        # using PN spins at omega_switch_IG
        quat_sur, orbphase_sur, chiA_copr_sur, chiB_copr_sur \
            = self.nrsur._sur_dimless.get_dynamics(q, chiA_PN_at_idx_coorb, \
            chiB_PN_at_idx_coorb, init_quat=quat_PN_copr_at_idx,
            init_orbphase=phi_PN_at_idx, omega_ref=omega_switch_IG)
        dyn_times = self.nrsur._sur_dimless.tds
        omega_sur = np.gradient(orbphase_sur, dyn_times)

        # Get surrogate orbital frequency at t_sur_switch, which is
        # close to the start of the surrogate data
        omega_init_sur = omega_sur[np.argmin(np.abs( \
                dyn_times - t_sur_switch))]

        # Get PN coorbital frame spins and frame dynamics at omega_init_sur
        idx = np.argmin(np.abs(omega_PN - omega_init_sur))
        chiA_PN_at_idx_coorb, chiB_PN_at_idx_coorb, quat_PN_copr_at_idx, \
            phi_PN_at_idx, omega_PN_at_idx \
            = self._get_coorbital_frame_spins_at_idx(chiA_PN, chiB_PN, \
            omega_PN, lNhat_PN, phi_PN, idx)

        return chiA_PN_at_idx_coorb, chiB_PN_at_idx_coorb, \
            quat_PN_copr_at_idx, phi_PN_at_idx, omega_PN_at_idx, \
            chiA_PN, chiB_PN, omega_PN


    #-------------------------------------------------------------------------
    def _evolve_spins(self, q, chiA0, chiB0, omega0, \
            return_spin_evolution=False, **kwargs):
        """ Evolves spins of the component BHs from an initial orbital
        frequency = omega0 until t=-100 M from the peak of the waveform.  If
        omega0 < omega_switch_IG, use PN to evolve the spins until
        t=t_sur_switch. Then evolves further with the NRSur7dq4 waveform model
        until t=-100M from the peak.

        Returns spins in the coorbital frame at t=-100M, as well as the
        coprecessing frame quaternion and orbital phase in the coprecessing
        frame at this time.

        If return_spin_evolution is given, also returns the PN and surrogate
        spin times series.
        """

        PN_approximant = kwargs.pop('PN_approximant', 'SpinTaylorT4')
        PN_dt = kwargs.pop('PN_dt', 0.1)
        PN_spin_order = kwargs.pop('PN_spin_order', 7)
        PN_phase_order = kwargs.pop('PN_phase_order', 7)
        # Initial guess for surrogate omega0, this should be large enough for
        # all q=6 cases
        omega_switch_IG = kwargs.pop('omega_switch_IG', 0.03)
        # The surrogate begins at -4300, use -4000 to be safe
        t_sur_switch = kwargs.pop('t_sur_switch', -4000)
        self._check_unused_kwargs(kwargs)

        # Load NRSur7dq4 if not previously loaded
        if self.nrsur is None:
            self._load_NRSur7dq4()

        # If omega0 is below the NRSur7dq4 initial guess frequency, we use PN
        # to evolve the spins. We get the initial spins and omega_init_sur such
        # that should go into the surrogate such that the inital time is
        # t_sur_switch.
        if omega0 < omega_switch_IG:
            chiA0_nrsur_coorb, chiB0_nrsur_coorb, quat0_nrsur_copr, \
                phi0_nrsur, omega_init_sur, chiA_PN, chiB_PN, omega_PN \
                = self._get_PN_spins_at_surrogate_start(PN_approximant, q, \
                omega0, chiA0, chiB0, PN_dt, PN_spin_order, PN_phase_order, \
                omega_switch_IG, t_sur_switch)

        # If omega0 >= omega_switch_IG, we evolve spins directly with NRSur7dq4
        # waveform model. We set the coprecessing frame quaternion to identity
        # and orbital phase to 0 at omega=omega0, hence the coprecessing frame
        # is the same as the inertial frame here.
        else:
            # Note that here we set omega_init_sur to omega0
            chiA0_nrsur_coorb, chiB0_nrsur_coorb, quat0_nrsur_copr, \
                phi0_nrsur, omega_init_sur, chiA_PN, chiB_PN, omega_PN \
                = chiA0, chiB0, [1,0,0,0], 0, omega0, None, None, None

        # Now evaluate the surrogate dynamics using PN spins at omega_init_sur
        quat_sur, orbphase_sur, chiA_copr_sur, chiB_copr_sur \
            = self.nrsur._sur_dimless.get_dynamics(q, chiA0_nrsur_coorb, \
            chiB0_nrsur_coorb, init_quat=quat0_nrsur_copr,
            init_orbphase=phi0_nrsur, omega_ref=omega_init_sur)

        # get data at time node where remnant fits are done
        dyn_times = self.nrsur._sur_dimless.tds
        nodeIdx = np.argmin(np.abs(dyn_times - self.fitnode_time))
        quat_fitnode = quat_sur.T[nodeIdx]
        orbphase_fitnode = orbphase_sur[nodeIdx]

        # get coorbital frame spins at the time node
        chiA_coorb_fitnode = utils.rotate_in_plane(chiA_copr_sur[nodeIdx],
                orbphase_fitnode)
        chiB_coorb_fitnode = utils.rotate_in_plane(chiB_copr_sur[nodeIdx],
                orbphase_fitnode)

        if return_spin_evolution:
            # Transform spins to the reference inertial frame
            chiA_inertial_sur = utils.transformTimeDependentVector(quat_sur, \
                    chiA_copr_sur.T).T
            chiB_inertial_sur = utils.transformTimeDependentVector(quat_sur, \
                    chiB_copr_sur.T).T
            spin_evolution = {
                    't_sur': dyn_times,
                    'chiA_sur': chiA_inertial_sur,
                    'chiB_sur': chiB_inertial_sur,
                    'orbphase_sur': orbphase_sur,
                    'quat_sur': quat_sur,
                    'omega_PN': omega_PN,
                    'chiA_PN': chiA_PN,
                    'chiB_PN': chiB_PN,
                    'omega_init_sur': omega_init_sur,
                }
        else:
            spin_evolution = None

        return chiA_coorb_fitnode, chiB_coorb_fitnode, quat_fitnode, \
            orbphase_fitnode, spin_evolution

    #-------------------------------------------------------------------------
    def _eval_wrapper(self, fit_key, q, chiA, chiB, **kwargs):
        """Evaluates the NRSur7dq4Remnant model.
        """
        chiA = np.array(chiA)
        chiB = np.array(chiB)

        # Warn/Exit if extrapolating
        allow_extrap = kwargs.pop('allow_extrap', False)
        omega0 = kwargs.pop('omega0', None)
        self._check_param_limits(q, chiA, chiB, allow_extrap)

        if omega0 is None:
            # If omega0 is given, assume chiA, chiB are the coorbital frame
            # spins at t=-100 M.
            x = np.concatenate(([q], chiA, chiB))
        else:
            # If omega0 is given, evolve the spins from omega0
            # to t = -100 M from the peak.
            chiA_coorb_fitnode, chiB_coorb_fitnode, quat_fitnode, \
                orbphase_fitnode, _ \
                = self._evolve_spins(q, chiA, chiB, omega0, **kwargs)

            # x should contain coorbital frame spins at t=-100M
            x = np.concatenate(([q], chiA_coorb_fitnode, chiB_coorb_fitnode))

        def eval_vector_fit(x, fit_key):
            res = self._evaluate_fits(x, fit_key)
            fit_val = res.T[0]
            fit_err = res.T[1]
            if omega0 is not None:
                # The fit are constructed in the coorbital frame at t=-100M,
                # now we transform the remnant vectors into the LAL inertial
                # frame, which is the same as the coorbital frame at omega0.
                fit_val = utils.transform_vector_coorb_to_inertial(fit_val,
                    orbphase_fitnode, quat_fitnode)
                fit_err = utils.transform_error_coorb_to_inertial(fit_val,
                    fit_err, orbphase_fitnode, quat_fitnode)
            return fit_val, fit_err


        if fit_key == 'mf' or fit_key == 'all':
            mf, mf_err = self._evaluate_fits(x, 'mf')
            if fit_key == 'mf':
                return mf, mf_err

        if fit_key == 'chif' or fit_key == 'all':
            chif, chif_err = eval_vector_fit(x, 'chif')
            if fit_key == 'chif':
                return chif, chif_err

        if fit_key == 'vf' or fit_key == 'all':
            vf, vf_err = eval_vector_fit(x, 'vf')
            if fit_key == 'vf':
                return vf, vf_err

        if fit_key == 'all':
            return mf, chif, vf, mf_err, chif_err, vf_err



import numpy as np
import h5py
import os
import warnings

import surfinBH

# Ignore these specific warnings.
warnings.filterwarnings("ignore", message="Mass ratio outside training range.")
warnings.filterwarnings("ignore", message="Spin magnitude of BhA outside"
        " training range.")
warnings.filterwarnings("ignore", message="Spin magnitude of BhB outside"
        " training range.")
warnings.filterwarnings("ignore", message="Extrapolating dynamics to")

def single_kwargs_test(fit, num_tests, kwargs={}):

    for i in range(num_tests):
        # Generate params randomly within allowed values
        q, chiA, chiB = fit._generate_random_params_for_tests() 

        # Check evaluation of different call modules

        # remnant mass
        mf_and_err = fit.mf(q, chiA, chiB, **kwargs)

        # remnant spin
        chif_and_err = fit.chif(q, chiA, chiB, **kwargs)

        # remnant kick
        vf_and_err = fit.vf(q, chiA, chiB, **kwargs)

        # all together
        #mf, chif, vf, mf_err_est, chif_err_est, vf_err_est \
        all_data = fit.all(q, chiA, chiB, **kwargs)

        # Check that all_data has the right ordering
        rtol = 1e-11

        if hasattr(mf_and_err, '__len__'):      # has both fit val and err_est
            np.testing.assert_allclose(mf_and_err[0], all_data[0], rtol=rtol)
            np.testing.assert_allclose(mf_and_err[1], all_data[3], rtol=rtol)
        else:
            np.testing.assert_allclose(mf_and_err, all_data[0], rtol=rtol)

        if hasattr(chif_and_err, '__len__'):      # has both fit val and err_est
            np.testing.assert_allclose(chif_and_err[0], all_data[1], rtol=rtol)
            np.testing.assert_allclose(chif_and_err[1], all_data[4], rtol=rtol)
        else:
            np.testing.assert_allclose(chif_and_err, all_data[1], rtol=rtol)

        if hasattr(vf_and_err, '__len__'):      # has both fit val and err_est
            np.testing.assert_allclose(vf_and_err[0], all_data[2], rtol=rtol)
            np.testing.assert_allclose(vf_and_err[1], all_data[5], rtol=rtol)
        else:
            np.testing.assert_allclose(vf_and_err, all_data[2], rtol=rtol)


def test_interface():
    """ Tests that the call modules are evaluating without breaking.
        Also tests that fit.all is doing the right thing.
    """
    # List of all available fits
    fit_names = surfinBH.fits_collection.keys()
    for name in fit_names:
        # Load fit
        fit = surfinBH.LoadFits(name)

        num_tests = 3

        # Test without any kwargs
        single_kwargs_test(fit, num_tests)

        # Test with different optional kwargs
        # NOTE: this would be empty unless _extra_regression_kwargs() is
        # overridden
        extra_kwargs_list = fit._extra_regression_kwargs()
        for i in range(len(extra_kwargs_list)):
            kwargs = extra_kwargs_list[i]
            single_kwargs_test(fit, num_tests, kwargs=kwargs)



import os
import errno
try:
    from urllib.request import urlretrieve # py 3
except ImportError:
    from urllib import  urlretrieve # py 2
from time import gmtime, strftime
from glob import glob

from . import _fit_evaluators
from ._dataPath import DataPath

#=============================================================================
class FitAttributes(object):
    """ Saves attributes of a particular fit.
    """
    #-------------------------------------------------------------------------
    def __init__(self, **kwargs):
        self.fit_class =  kwargs['fit_class']
        self.desc = kwargs['desc']
        self.data_url =  kwargs['data_url']
        self.refs =  kwargs['refs']

#-------------------------------------------------------------------------
def LoadFits(name):
    """ Loads data for a fit.
    If data is not available, downloads it before loading.
    """
    if name not in fits_collection.keys():
        raise Exception('Invalid fit name : %s'%name)
    else:
        testPath = DataPath() + '/' + fits_collection[name].data_url.split('/')[-1]
        if (not os.path.isfile(testPath)):
            DownloadData(name)

        # allow for both naming formats surfinBH7dq2 and NRSur7dq4Remnant
        if 'surfinBH' in name:
            name_tag = name.split('surfinBH')[-1]
        else:
            name_tag = name.split('NRSur')[-1].split('Remnant')[0]

        fit = fits_collection[name].fit_class(name_tag)
        print('Loaded %s fit.'%name)
        return fit

#-------------------------------------------------------------------------
def DownloadData(name='all', data_dir=DataPath()):
    """ Downloads fit data to DataPath() diretory.
        If name='all', gets all fit data.
    """
    if name == 'all':
        for tmp_name in fits_collection.keys():
            DownloadData(name=tmp_name, data_dir=data_dir)
        return

    if name not in fits_collection.keys():
        raise Exception('Invalid fit name : %s'%name)

    print('Downloading %s data'%name)
    data_url = fits_collection[name].data_url
    filename = data_url.split('/')[-1]
    try:
        os.makedirs(data_dir)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(data_dir):
            pass
        else:
            raise
    urlretrieve(data_url, data_dir + '/' + filename)


##############################################################################

#### Add fits here, each name should start with surfinBH.
fits_collection = {}

fits_collection['NRSur3dq8Remnant'] = FitAttributes( \
    fit_class = _fit_evaluators.Fit3dq8,
    desc = 'Fits for remnant mass, spin and kick veclocity for nonprecessing'
        ' BBH systems. This model was called surfinBH3dq8 in the paper.',
    data_url = 'https://zenodo.org/record/1435751/files/remnant_fits/fit_3dq8.h5',
    refs = 'arxiv:1809.09125',
    )

fits_collection['surfinBH7dq2'] = FitAttributes( \
    fit_class = _fit_evaluators.Fit7dq2,
    desc = 'Fits for remnant mass, spin and kick veclocity for generically'
        ' precessing BBH systems.',
    data_url = 'https://zenodo.org/record/1435751/files/remnant_fits/fit_7dq2.h5',
    refs = 'arxiv:1809.09125',
    )


fits_collection['NRSur7dq4Remnant'] = FitAttributes( \
    fit_class = _fit_evaluators.Fit7dq4,
    desc = 'Fits for remnant mass, spin and kick veclocity for generically'
        ' precessing BBH systems up to mass ratio 4.',
    data_url = 'https://zenodo.org/record/3455886/files/remnant_fits/fit_7dq4.h5',
    refs = 'arxiv:1905.09300',
    )
